#include <FunctionHook/Hook.h>
#include <sys/mman.h>
#include <csignal>
#include <cstring>
#include <array>

using namespace FunctionHooking;
constexpr std::array<uint8_t, 13> empty_hook = { 0x49, 0xBA, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x41, 0xFF, 0xE2 };

void Detour::CreateHook(void* source_address, void* destination_address) {
    hook_bytes.resize(empty_hook.size());
    memcpy(hook_bytes.data(), empty_hook.data(), hook_bytes.size());

    overwritten_bytes.resize(hook_bytes.size());

    //Save the bytes to be overwritten from the function prelude.
    memcpy(overwritten_bytes.data(), source_address, hook_bytes.size());

    //Put the destination address into the jump.
    memcpy(&hook_bytes[2], &destination_address, sizeof(void*));

    //Make the target memory page writable.
    mprotect((void*)((uintptr_t)source_address & ~(sysconf(_SC_PAGE_SIZE)-1)), sysconf(_SC_PAGE_SIZE), PROT_READ|PROT_WRITE|PROT_EXEC);

    //Write the jmp to the beginning of the function prelude.
    memcpy(source_address, hook_bytes.data(), hook_bytes.size());

    //Make the memory page non-writable again.
    mprotect((void*)((uintptr_t)source_address & ~(sysconf(_SC_PAGE_SIZE)-1)), sysconf(_SC_PAGE_SIZE), PROT_READ|PROT_EXEC);
}

bool Detour::Hooked(void* function_address) {
    std::vector<uint8_t> current_data(empty_hook.size());
    memcpy(current_data.data(), function_address, current_data.size());

    if (current_data[0] == empty_hook[0] && current_data[1] == empty_hook[1] && current_data[10] == empty_hook[10]
    && current_data[11] == empty_hook[11] && current_data[12] == empty_hook[12])
        return true;
    return false;
}

void VTable::CreateHook(void* class_inst, int offset, void* dest) {
    class_instance = class_inst;
    destination = dest;
    vtable_offset = offset;
    intptr_t vtable = *((intptr_t*) class_inst);
    intptr_t entry = vtable + sizeof(intptr_t) * offset;
    original = (intptr_t*) *((intptr_t*) entry);

    mprotect((void*) ((uintptr_t) entry & ~(sysconf(_SC_PAGE_SIZE)-1)), sysconf(_SC_PAGE_SIZE), PROT_READ| PROT_WRITE| PROT_EXEC);
    *((intptr_t *) entry) = (intptr_t) dest;
    mprotect((void*) ((uintptr_t) entry & ~(sysconf(_SC_PAGE_SIZE)-1)), sysconf(_SC_PAGE_SIZE), PROT_READ | PROT_EXEC);
}

void VTable::RemoveHook() {
    intptr_t vtable = *((intptr_t*) class_instance);
    intptr_t entry = vtable + sizeof(intptr_t) * vtable_offset;

    mprotect((void*) ((uintptr_t) entry & ~(sysconf(_SC_PAGE_SIZE)-1)), sysconf(_SC_PAGE_SIZE), PROT_READ| PROT_WRITE| PROT_EXEC);
    *((intptr_t *) entry) = (intptr_t) original;
    mprotect((void*) ((uintptr_t) entry & ~(sysconf(_SC_PAGE_SIZE)-1)), sysconf(_SC_PAGE_SIZE), PROT_READ | PROT_EXEC);
}

VTable::VTable(void* class_instance, int vtable_offset, void* destination) {
    CreateHook(class_instance, vtable_offset, destination);
}

VTable::~VTable() {
    if (Valid())
        RemoveHook();
}

bool VTable::Valid() {
    intptr_t vtable = *((intptr_t*) class_instance);
    intptr_t entry = vtable + sizeof(intptr_t) * vtable_offset;
    return *((intptr_t*) entry) == (intptr_t) destination;
}

bool VTable::Hooked(void* class_instance_to_check, void* class_instance_to_compare_against) {
    return *reinterpret_cast<void**>(class_instance_to_check) != *reinterpret_cast<void**>(class_instance_to_compare_against);
}
